//-----------------------------------------------------------------
// Esempio applicativo del protocollo MassaBus con Arduino
// Il programma per la prova di funzionamento richiede una scheda
// di output (Multi I/O o Mini I/O con indirizzo pre impostato a 2)
// Durante la prova ogni circa 5 secondi l'output sulla scheda cambierà
// di stato
//------------------------------------------------------------------

//Definisco numero byte bus
#define SIZE_BUS_DATA 11
//Definisco il timeout massimo per la seriale
#define MAX_MILLIS_TIMEOUT 50

//Inizializzazioni
int led_act = 13; //Pin di arduino con il LED incluso
int bus_dir = 2; //Pin di selezione della direzione
int count; //Variabile conteggio generica

long mil_mem = 0; //Variabile memorizzazione millisecondi

byte my_adr = 1; //Mio indirizzo nel bus
byte send_compose[SIZE_BUS_DATA]; //Array invio pacchetto

//Variabili procedure ricezione (riguardano l'ultimo pacchetto decodificato)
byte recived_data[SIZE_BUS_DATA]; //Array del pacchetto ricevuto
boolean correct_data; //True se il pacchetto era corretto
byte adress_reciver_ric; //L'indirizzo destinatario
byte adress_sender_ric; //Indirizzo mittente
byte type_ric; //Tipologia pacchetto
byte data1_ric; //Campo dato 1
byte data2_ric; //Campo dato 2

//Procedura creazione trama da inviare su BUS
//Campi: indirizzo destinatario,tipologia pacchetto, dato 1, dato 2)
void code_data_to_send(byte adress_reciver, byte type_s, byte data1_s, byte data2_s)
{
 //Variabili utilizzate nella procedura
byte checksum_send; //checksum completo
byte checksum_send_low; //checksum diviso (parte bassa)
byte checksum_send_hi; //checksum diviso (parte alta)
byte type_to_send; //Tipologia pacchetto
byte data_send_1; //Primo campo dati da inviare
byte data_send_2; //Secondo campo dati da inviare
byte adress_sender; //Indirizzo mittente
//Inizio procedura
type_to_send = type_s << 2; //shift di due bit per spazio riservato zero flag
//se l'indirizzo è in modalità normale (destinatario diverso da 255)
if(adress_reciver != 255){
  adress_sender = my_adr; //Metto come mittente il mio indirizzo
 } else {
  adress_sender = 255; //Altrimenti anche il mittente a 255
 } 
if(data1_s == 0){ //se data1 è pari a 0
  data_send_1 = 48; //lo sostituisco con il codice ASCII 48
  type_to_send = type_to_send | 1; //Imposto a 1 il rispettivo zero flag
 } else {
  data_send_1 = data1_s; //altrimenti mantengo il dato
 }
if(data2_s == 0){ //se data2 è pari a 0
  data_send_2 = 48; //lo sostituisco con il codice ASCII 48
  type_to_send = type_to_send | 2; //Imposto a 1 il rispettivo zero flag
 } else {
  data_send_2 = data2_s; //altrimenti mantengo il dato
 }
//Calcolo e assegnazione bit checksum
//Effettuo XOR con tutti i campi necessari
checksum_send = adress_reciver ^ adress_sender ^ type_to_send ^ data_send_1 ^ data_send_2;
//Suddivido il checksum in due parti (alta e bassa) a 4 bit ciascuna
//Imposto di ogni parte il bit più significativo a 1 come previsto da protocollo
checksum_send_low = (checksum_send & 15) | 128;
checksum_send_hi = (checksum_send >> 4) | 128;
// Composizione array stringa seriale
//Ad ogni rispettivo byte del pacchetto assegno il rispettivo dato
send_compose[0] = 'S';
send_compose[1] = adress_reciver;
send_compose[2] = adress_sender;
send_compose[3] = type_to_send;
send_compose[4] = data_send_1;
send_compose[5] = data_send_2;
send_compose[6] = checksum_send_low;
send_compose[7] = checksum_send_hi;
send_compose[8] = 3; //Terminazione
send_compose[9] = 3;
send_compose[10] = 3;
//Invio il pacchetto
send_data();

}

//Procedura invio dati su bus tramite seriale
void send_data(){
 // setto il flag di trasmissione
 UCSR0A=UCSR0A |(1 << TXC0);
 Serial.flush();
 // abilito led
 digitalWrite(led_act, HIGH);
 //attivo il MAX485 in trasmissione
 digitalWrite(bus_dir, HIGH);
 //Mando l'array di byte
 Serial.write(send_compose, SIZE_BUS_DATA);
 //Attendo la fine della trasmissione
 Serial.flush();
 while (!(UCSR0A & (1 << TXC0)));
 //Disabilito led e max485
 digitalWrite(bus_dir, LOW); 
 digitalWrite(led_act, LOW);
}

//Procedura decodifica dati

void decode_ric_data(){
//Variabili checksum separati
byte chk_low;
byte chk_hi;
//Composizione complessiva del checksum
byte checksum;
//Variabile checksum ricalcolato
byte calcolated_checksum;

//imposto a default la variaible pacchetto corretto
correct_data = false;
//Verifico che il primo pacchetto sia S e ci sia la terminazione (come da protocollo MassaBUS)
if((recived_data[0] == 'S')&&(recived_data[8] == 3)){ //Controllo start byte e terminazione
 //Ricostruzione byte checksum
 //Separo la parte bassa dal pacchetto
 chk_low = recived_data[6] & 15;
 //Separo la parte alta dal pacchetto
 chk_hi = recived_data[7] & 15;
 //Mette insieme i due byte per ricostrire checksum (con shift del più alto)
 checksum = (chk_hi << 4) | chk_low; 
 //controllo il dato ricevuto con il checksum (lo ricostruisco con i byte ricevuti)
 calcolated_checksum = recived_data[1] ^ recived_data[2] ^ recived_data[3] ^ recived_data[4] ^ recived_data[5];
 //Verifico che il checksum sia uguale (calcolato e ricevuto) e verifico che i due byte dei
 //pacchetti dei checksum abbiano il bit più significativo alto (come da protocollo)
 if((calcolated_checksum == checksum) && ((recived_data[6] & 128) == 128) && ((recived_data[7] & 128) == 128)){
   //se il checksum è corretto
   //Setto nella variabile condivisa l'indirizzo del destinatario (forse questa scheda!!!)
   adress_reciver_ric = recived_data[1];
   //Setto nella variabile condivisa l'indirizzo del mittente
   adress_sender_ric =recived_data[2];
   //preleva il pacchetto di tipologia (saltando gli zero flag)
   type_ric = recived_data[3] >> 2; 
   //Verifica degli zero flag (sia primo che secondo campo dati)
   //Se dato 48 e rispettivo zero flag a 1 setto data a 0
   //altrimenti setto il valore ricevuto
   if((recived_data[4]==48)&&((recived_data[3] & 1)==1)){
     //se data1 = 0 (controllo zero flag a 1)
     data1_ric=0;
   } else {
     data1_ric=recived_data[4];
   }
   //Verifica degli zero flag
   //Se dato 48 e rispettivo zero flag a 1 setto data a 0
   //altrimenti setto il valore ricevuto
   if((recived_data[5]==48)&&((recived_data[3] & 2)==2)){
     //se data2 = 0 (controllo zero flag a 1)
     data2_ric=0;
   } else {
     data2_ric=recived_data[5];
   }
   //Indico che il pacchetto è stato decodificato correttamente
   correct_data = true;
 }
}
 //Se il pacchetto è decodificato correttamente
 if(correct_data){
   //Chiamo la procedura che gestisce il pacchetto decodificato
   packet_data_elaboration();
 }
}


// the setup routine runs once when you press reset:
void setup() {                
  // inizializzo led indicazione invio pacchetto e led selezione direzione
  pinMode(led_act, OUTPUT);  
  pinMode(bus_dir, OUTPUT);   
  Serial.begin(9600);
}

//evento lanciato durante la ricezione di dati seriali
void serialEvent() {
  byte index_buffer_rx = 0;
  boolean get_byte_continue = true;
  //Prendo i millisecondi attuali
  long last_millis = millis();
  long mem_millis;
  //Inizializzo conteggio millisecondi di timeout
  int count_millis = 0;
  //Inizializzo varibile indicante dati corretti
  boolean bus_data_avaible = false;
  //Verifico disponibilità dati
  if(Serial.available()){
    //Prelevo primo dato
    recived_data[index_buffer_rx++] = Serial.read(); 
    //Attivo loop di ricezione pacchetti
    while(get_byte_continue){
      //Parte gestione byte ricevuti
      if(Serial.available()){
        recived_data[index_buffer_rx++] = Serial.read(); 
        //Verifico se il buffer ha un pacchetto completo
        // (-2 per riconosere il pacchetto conteggiando da 0)
        if(index_buffer_rx > (SIZE_BUS_DATA - 2)){
          //Fermo la ricezione di byte (dopo uscirà dal ciclo)
          get_byte_continue = false;
          //Indico mettendo in true che il pacchetto è stato ricevuto correttamente
          bus_data_avaible = true;
        }
      }
      //Seconda parte (conteggia i milliscondi del ciclo while)
      //Se supera timeout esco (evitando stalli del software!)
      //Parte gestione timeout
      //Prelevo i millisecondi attuali dall'accensione di arduino
      mem_millis = millis();
      //Se diversi da quelli memorizzati precedentemente
      if(mem_millis != last_millis){
        //Salvo in memoria gli attuali
        last_millis = mem_millis;
        //Incremento il conteggio dei millisecondi
        count_millis++;
        //Se supera soglia di millisecondi impostati (costante)
        if(count_millis > (MAX_MILLIS_TIMEOUT - 1)){
          //Esco dal ciclo mettendo in flase la variabile
          get_byte_continue = false;
          //bus_data_avaible rimane in false (pacchetto non corretto!!)
        }
      }
     }
  }
  //Se all'uscita del while i dati sono corretti
  if(bus_data_avaible){
    //Chiamo la procedura di decodifica pacchetto
    //Nell'array recived_data ho tutti i byte del pacchetto
    //Pronti da analizzae...
    decode_ric_data();  
  }
}

void loop() { 
  //Attendo 5 secondi (non uso delay per permettere ricezione seriale)
 if (millis() != mil_mem){
   mil_mem = millis();
 if(count++ > 5000){
   count = 0;
   //Invio in output 5 (tipologia 12 output, indirizzo 2)
 code_data_to_send(2, 12, 5, 0);
 }
}
}
 
void packet_data_elaboration() {
//Se il pacchetto è indirizzato qui e arriva dal dispositivo 2 con tipologia 12
if ((adress_reciver_ric == 1) && (adress_sender_ric == 2) && (type_ric == 12)){
  if(data1_ric = 5){
    delay(5000); //attendo 5 sec
    code_data_to_send(2, 12, 10, 0);
    //Invio in output 10 (tipologia 12 output, indirizzo 2)
  }   
 }
}
  
